#include "gd.h"

// #define GD_SERIAL_DEBUG

uint8_t GDdisplayedBank; // 0 or 1
uint8_t GDcurrentBank; // 0 or 1 - bank currently selected for drawing
uint8_t GDgraphicsDisplayed; // 0 or 1
uint8_t GDleftByte, GDrightByte; // for doing 16-bit read/writes from/to display memory
// following two bytes keep track of whether read/write of the display memory is needed - for example multiple read/writes within the same 16 pixels can be cached
// this assumes that the pixels are being written to the off-screen buffer: an immediate write occurs when GDdisplayedBank == GDcurrentBank
// because the Xaddr is always in the range 0-15, two bits of GDprevXaddr are used for 'address changed' (0x10) and 'contents altered' (0x20) flags
uint8_t GDprevYaddr, GDprevXaddr; 

uint8_t GDreadDataPort(void) {
  // datasheet says (at Ta=25 degreesC, Vdd=4.5V) cycle time from previous E_HIGH Tc >1200ns = 29 cycles with 16MHz clock
  PORTD &= 0x03; // pins 2 - 7 no pullup
  DDRD &= 0x03; // pins 2 - 7 input
  PORTB &= 0xFC; // pins 8 - 9 no pullup
  DDRB &= 0xFC; // pins 8 - 9 input
  GD_RW_HIGH;
  GD_E_HIGH; // datasheet says (at Ta=25 degreesC, Vdd=4.5V) Enable pulse width Tpw >140ns = 3 cycles with 16MHz clock
  asm("nop \n");  
  asm("nop \n");  
  asm("nop \n");  
  uint8_t b = (PIND >> 2) | (PINB << 6);
  GD_E_LOW;
  return b;  
}

void GDwriteDataPort(uint8_t b) {
  // datasheet says (at Ta=25 degreesC, Vdd=4.5V) cycle time from previous E_HIGH Tc >1200ns = 19 cycles with 16MHz clock
  GD_E_HIGH;
  DDRD |= 0xFC; // pins 2 - 7 output
  DDRB |= 0x03; // pins 8 - 9 output
  PORTD = b << 2; // data bus bits span 2 ports: write the low 6 bits to PORTD
  if (b & 0b01000000) { // now write the upper 2 bits to PORTB without disturbing PORTB's other bits
    asm("sbi 0x05, 0 \n");
  } else {
    asm("cbi 0x05, 0 \n");
  }
  if (b & 0b10000000) {
    asm("sbi 0x05, 1 \n");
  } else {
    asm("cbi 0x05, 1 \n");
  }
  // datasheet says (at Ta=25 degreesC, Vdd=4.5V) Enable pulse width Tpw >140ns = 3 cycles with 16MHz clock
  GD_E_LOW;
}

void GDwaitWhileBusy() {
  // read busy flag (also gets address counter AC)
  GD_RS_LOW;
  while (GDreadDataPort() & 0x80) { // loop while display busy
#ifdef GD_SERIAL_DEBUG
    Serial.print('.');
#endif    
  }  
}

void GDramWrite(uint8_t b) {
  GDwaitWhileBusy();
#ifdef GD_SERIAL_DEBUG
  Serial.print("\n=>0x");
  if (b < 0x10) {
    Serial.print('0');
  }
  Serial.print(b, HEX);
#endif
  GD_RS_HIGH;
  GD_RW_LOW;
  GDwriteDataPort(b);  
}

void GDinsWrite(uint8_t b) {
  GDwaitWhileBusy();
#ifdef GD_SERIAL_DEBUG
  Serial.print("\n=>0x");
  if (b < 0x10) {
    Serial.print('0');
  }
  Serial.print(b, HEX);
#endif
  GD_RW_LOW;
  GDwriteDataPort(b);
}

void GDgClear(void) {
  GDinsWrite(GDgraphicsDisplayed ? 0x36 : 0x34); // function set RE for following extended instruction
  GDinsWrite(0b00000010); // select IRAM or CGRAM address set (not vertical scroll address)
  uint8_t rowStart = GDcurrentBank ? 32 : 0;
  uint8_t rowFinish = GDcurrentBank ? 64 : 32;
  for (uint8_t row = rowStart; row < rowFinish; row++) {
    GDinsWrite(0b10000000 | row); // set graphic RAM row
    GDinsWrite(0b10000000 | 0); // set graphic column (0) (16 pixels across each column)
    for (uint8_t i = 0; i < 16; i++) { // clear one row of pixels first half of 'row' on top half of buffer, second half of 'row' on bottom half of buffer
      GDramWrite(0x00);
      GDramWrite(0x00);
    }
  }
  GDinsWrite(GDgraphicsDisplayed ? 0x32 : 0x30); // extended function set, back to basic instruction set
  GDprevXaddr = 0x10; // set 'addressed changed' flag, clear 'contents altered' flag 
}

void GDinit(void) {
  GD_RS_LOW;
  GD_RW_HIGH;
  GD_E_LOW;
  DDRB |= 0x1C; // pins 10, 11 and 12 output (RS, RW, and E)
  delay(40);
  GDinsWrite(0x30); // function set 8-bit interface
  GDinsWrite(0x34); // function set RE for following extended instruction
  GDinsWrite(0x34); // extended instruction: graphic display off
  GDgraphicsDisplayed = 0;
  GDinsWrite(0x30); // extended function set, back to basic instruction set (can't change G and RE flags at same time)
  GDinsWrite(0x84); // display on, cursor off, blink off
  GDinsWrite(0x01); // clear DDRAM (text) and set auto-increment (not decrement) after write or read
  GDinsWrite(0x06); // no display shift after write or read
  GDcurrentBank = 0;
  GDgClear(); // clear (currently hidden) graphics memory bank 0
  GDcurrentBank = 1; // double buffering initialized with bank 1 selected for writing, and bank 0 displayed
  GDgClear(); // clear (currently hidden and not selected for display) graphics memory bank 1
  GDdisplayedBank = 0; // the 0x01 clear DDRAM instruction auto-selects bank 0 for display
  GDprevXaddr = 0x10; // set 'addressed changed' flag, clear 'contents altered' flag   
}

void GDflushGraphicsPixelBuffer(void) {
  if (GDprevXaddr & 0x20) { // contents altered, write the two changed bytes
#ifdef GD_SERIAL_DEBUG
    Serial.print("*->(");
    Serial.print(GDprevXaddr & 0x0F);
    Serial.print(',');
    Serial.print(GDprevYaddr & 0x7F);
    Serial.print(')');
#endif
    GDramWrite(GDleftByte);
    GDramWrite(GDrightByte);
    GDprevXaddr |= 0x10; // set 'address changed' flag
    GDprevXaddr &= ~0x20; // clear 'contents altered' flag
  }
}

void GDsetDDaddress(uint8_t addr) {
  GDflushGraphicsPixelBuffer();
  if (GDcurrentBank) {
    GDinsWrite(0xA0 | addr);
  } else {
    GDinsWrite(0x80 | addr);
  }
}

uint8_t GDreadRam(void) {
  GDwaitWhileBusy();
  GD_RS_HIGH;
  return GDreadDataPort();  
}

#ifdef GD_SERIAL_DEBUG
void GDhexPrint(uint8_t b) {
  if (b < 0x10) Serial.print(' ');
  Serial.print(b, HEX);
  Serial.print(' ');
}
#endif

void GDread16bits(void) {
  GDleftByte = GDreadRam();
  GDrightByte = GDreadRam();
#ifdef GD_SERIAL_DEBUG
  Serial.print("\n<=");
  GDhexPrint(GDleftByte);
  GDhexPrint(GDrightByte);
#endif  
}

void GDtWrite(uint8_t *text, uint8_t row, uint8_t col) {
  if (!*text) return; // null string
  if (row > 3) row = 3;
  if (col > 15) col = 15;
  uint8_t addr = col >> 1; // display addresses are per 16 pixels: 2 ASCII chars
  if (row & 0x01) addr += 16;
  if (row > 1) addr += 8;
  GDsetDDaddress(addr);
  uint8_t c1;
  if (col & 0x01) { // get existing first character of 2-character group from display
    GDreadRam(); // a dummy read is required here, because the port hasn't been read or written since setting the address
    GDread16bits();
    c1 = GDleftByte;
    GDsetDDaddress(addr); // read operation advances address, so wind back
  } else { // start of text not second character of a 2-character group
    c1 = *text++;
  }
  while (addr < 32) { // clip text that overflows end of currentBank
    uint8_t c2 = *text;
    if (!c2) { // get existing second character of 2-character group from display
      GDsetDDaddress(addr); // reset address because we don't know here whether or not port has already been read or written since previous address setting
      GDreadRam(); // dummy read following address set
      GDread16bits();
      c2 = GDrightByte;
      GDsetDDaddress(addr); // read operation advances address, so wind back
    } else {
      text++;
    }
    GDramWrite(c1);
    GDramWrite(c2);
    c1 = *text++;
    if (!c1) break; // reached end of string
    addr++;
  }
}

void GDtCopyBank(void) { // copy GDdisplayedBank to GDcurrentBank
  uint8_t currentBankCopy = GDcurrentBank;
  for (uint8_t addr = 0; addr < 32; addr++) {
    GDcurrentBank = GDdisplayedBank; 
    GDsetDDaddress(addr); // set address in GDdisplayed bank
    GDreadRam(); // dummy read following address set
    GDread16bits(); // read two bytes
    GDcurrentBank = currentBankCopy;
    GDsetDDaddress(addr); // set same address in GDcurrentBank
    GDramWrite(GDleftByte); // write first byte. No need for dummy opteration when writing
    GDramWrite(GDrightByte); // write the second byte
  }
}

void GDtClear(void) {
  GDtWrite("  ", 0, 0); // address GDcurrentBank and clear first 2 characters
  for (uint8_t i = 0; i < 62; i++) { // clear remaining 62 characters
    GDramWrite(' ');
  }
}

void GDflipBanks(void) {
  GDflushGraphicsPixelBuffer();
  GDcurrentBank ^= 1;
  GDdisplayedBank ^= 1;
  GDinsWrite(GDgraphicsDisplayed ? 0x36 : 0x34); // function set RE for following extended instructions
  GDinsWrite(0x03); // extended instruction: select vertical scroll address set
  GDinsWrite(GDdisplayedBank ? 0x60 : 0x40); // extended instruction: set scroll address to 0 for displayedBank 0 or 32 for displayedBank 1
  GDinsWrite(GDgraphicsDisplayed ? 0x32 : 0x30); // extended function set, back to basic instruction set
}

void GDdoubleBuffering(bool on) {
  if (on) {
    GDcurrentBank = GDdisplayedBank ^ 1;
  } else {
    GDcurrentBank = GDdisplayedBank;
  }
}

void GDgDisplay(bool graphicsOn) { // turn graphics display on or off
  GDgraphicsDisplayed = graphicsOn;
  GDinsWrite(GDgraphicsDisplayed ? 0x36 : 0x34); // function set RE for following extended instruction
  GDinsWrite(GDgraphicsDisplayed ? 0x36 : 0x34); // extended function set, turn graphics flag on or off
  GDinsWrite(GDgraphicsDisplayed ? 0x32 : 0x30); // extended function set, back to basic instruction set
}

void GDaddressPixel(uint8_t x, uint8_t y) { // read current pair of bytes containing pixel, then readdress for (probable) following write
  x >>= 4; // 16-bit addresses
  if (y & 0x20) { // bottom half of display is (from an addressing viewpoint) to the right of the top half
    x |= 0x08;
    y &= 0x1F;
  }
  y |= GDcurrentBank ? 0xA0 : 0x80; // y now sets CGRAM row of currentBank
  x |= 0x80; // x now sets CGRAM column
  if ((y != GDprevYaddr) || (x != (GDprevXaddr & 0x9F))) { // address changed
    GDflushGraphicsPixelBuffer();
    // now read and readdress the new location
    GDinsWrite(GDgraphicsDisplayed ? 0x36 : 0x34); // function set RE for following extended instructions
    GDinsWrite(0x02); // select IRAM or CGRAM address set (not vertical scroll address)
    GDinsWrite(y); // set CGRAM row  
    GDinsWrite(x); // set CGRAM column
    GDreadRam(); // dummy read following address set
    GDread16bits(); // read two bytes
    GDinsWrite(y); // set CGRAM row  
    GDinsWrite(x); // set CGRAM column  
    GDinsWrite(GDgraphicsDisplayed ? 0x32 : 0x30); // extended function set, back to basic instruction set
    GDprevYaddr = y;
    GDprevXaddr = x; // with 'address changed' and 'contents altered' flags clear
  }
}

void GDgWritePixel(uint8_t x, uint8_t y, uint8_t mode) {
  if (x & 0x80 || y & 0xC0) return; // pixel off screen
  GDaddressPixel(x, y);
  uint8_t * rightOrLeft = x & 0x08 ? &GDrightByte : &GDleftByte;
  uint8_t pixelMask = 0x80 >> (x & 0x07);
  switch (mode) {
    case 0: // clear pixel
      *rightOrLeft &= ~pixelMask;
      break;
    case 1: // set pixel
      *rightOrLeft |= pixelMask;
      break;
    default: // flip pixel
      *rightOrLeft ^= pixelMask;
      break;
  }
  if (((GDprevXaddr & 0x10) == 0x00) && (GDdisplayedBank != GDcurrentBank)) { // skip writing ram for now - wait till address changes, or it's on-screen
    GDprevXaddr |= 0x20; // set 'contents altered' flag
  } else {
    GDflushGraphicsPixelBuffer();
  }
}

bool GDgReadPixel(uint8_t x, uint8_t y) {
  if (x & 0x80 || y & 0xC0) return false; // pixel off screen
  GDaddressPixel(x, y);
  uint8_t * rightOrLeft = x & 0x08 ? &GDrightByte : &GDleftByte;
  uint8_t pixelMask = 0x80 >> (x & 0x07);
  return (*rightOrLeft & pixelMask) != 0x00;
}
